@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2003, 2004 by Juliusz Chroboczek
+Copyright (c) 2003-2006 by Juliusz Chroboczek
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -110,7 +110,7 @@ lwr(char a)
 }
 
 char *
-lwrcpy(char *dst, const char *src, int n)
+lwrcpy(char *restrict dst, const char *restrict src, int n)
 {
     int i;
     for(i = 0; i < n; i++)
@@ -209,6 +209,17 @@ h2i(char h)
         return -1;
 }
     
+char
+i2h(int i)
+{
+    if(i < 0 || i >= 16)
+        return '?';
+    if(i < 10)
+        return i + '0';
+    else
+        return i - 10 + 'A';
+}
+
 /* floor(log2(x)) */
 int
 log2_floor(int x) 
@@ -320,12 +331,15 @@ char *
 pstrerror(int e)
 {
     char *s;
+    static char buf[20];
+
     switch(e) {
     case EDOSHUTDOWN: s = "Immediate shutdown requested"; break;
     case EDOGRACEFUL: s = "Graceful shutdown requested"; break;
     case EDOTIMEOUT: s = "Timeout"; break;
     case ECLIENTRESET: s = "Connection reset by client"; break;
     case ESYNTAX: s = "Incorrect syntax"; break;
+    case EREDIRECTOR: s = "Redirector error"; break;
     case EDNS_HOST_NOT_FOUND: s = "Host not found"; break;
     case EDNS_NO_ADDRESS: s = "No address"; break;
     case EDNS_NO_RECOVERY: s = "Permanent name server failure"; break;
@@ -334,11 +348,44 @@ pstrerror(int e)
     case EDNS_UNSUPPORTED: s = "Unsupported DNS reply"; break;
     case EDNS_FORMAT: s = "Invalid DNS query"; break;
     case EDNS_REFUSED: s = "DNS query refused by server"; break;
+    case EDNS_CNAME_LOOP: s = "DNS CNAME loop"; break;
+#ifndef NO_SOCKS
+    case ESOCKS_PROTOCOL: s = "SOCKS protocol error"; break;
+    case ESOCKS_REJECT_FAIL: s = "SOCKS request rejected or failed"; break;
+    case ESOCKS_REJECT_IDENTD: s = "SOCKS request rejected: "
+                                   "server couldn't connect to identd";
+    case ESOCKS_REJECT_UID_MISMATCH: s = "SOCKS request rejected: "
+                                         "uid mismatch";
+        break;
+    case ESOCKS5_BASE: s = "SOCKS success"; break;
+    case ESOCKS5_BASE + 1: s = "General SOCKS server failure"; break;
+    case ESOCKS5_BASE + 2: s = "SOCKS connection not allowed"; break;
+    case ESOCKS5_BASE + 3: s = "SOCKS error: network unreachable"; break;
+    case ESOCKS5_BASE + 4: s = "SOCKS error: host unreachable"; break;
+    case ESOCKS5_BASE + 5: s = "SOCKS error: connection refused"; break;
+    case ESOCKS5_BASE + 6: s = "SOCKS error: TTL expired"; break;
+    case ESOCKS5_BASE + 7: s = "SOCKS command not supported"; break;
+    case ESOCKS5_BASE + 8: s = "SOCKS error: address type not supported";
+        break;
+#endif
     case EUNKNOWN: s = "Unknown error"; break;
     default: s = NULL; break;
     }
     if(!s) s = strerror(e);
-    if(!s) s = "Unknown error";
+#ifdef MINGW
+    if(!s) {
+        if(e >= WSABASEERR && e <= WSABASEERR + 2000) {
+            /* This should be okay, as long as the caller discards the
+               pointer before another error occurs. */
+            snprintf(buf, 20, "Winsock error %d", e);
+            s = buf;
+        }
+    }
+#endif
+    if(!s) {
+        snprintf(buf, 20, "Unknown error %d", e);
+        s = buf;
+    }
     return s;
 }
 
@@ -427,7 +474,6 @@ expandTilde(AtomPtr filename)
     
     home = getenv("HOME");
     if(home == NULL) {
-        do_log(L_ERROR, "Could not expand $HOME.\n");
         return NULL;
     }
     len = strlen(home);
@@ -449,6 +495,7 @@ expandTilde(AtomPtr filename)
     return ret;
 }
 
+#ifdef HAVE_FORK
 void
 do_daemonise(int noclose)
 {
@@ -478,6 +525,17 @@ do_daemonise(int noclose)
     }
 }
 
+#else
+
+void
+do_daemonise(int noclose)
+{
+    do_log(L_ERROR, "Cannot daemonise on this platform");
+    exit(1);
+}
+#endif
+
+
 void
 writePid(char *pidfile)
 {
@@ -543,7 +601,7 @@ b64cpy(char *restrict dst, const char *restrict src, int n, int fss)
 }
 
 int
-b64cmp(const char *a, int an, const char *b, int bn)
+b64cmp(const char *restrict a, int an, const char *restrict b, int bn)
 {
     char *buf;
     int r;
